#ifndef __COMMON_CLASSES__
#define __COMMON_CLASSES__

#include <fifo.h>
#include <qstring.h>

#define NUMBER		6
#define THIS		5

class OutputInfo;
class Value;
class DelayedValue;

// equation information for the Equations dialog
class OutputInfo
{
public:
	OutputInfo() { device_id = 0; output_id = 0; };
	~OutputInfo() {};

	int device_id;
	int output_id;
	QString equation;
	QString output_name;	// output name
	QString original_output_name;

	QString getOutputName() {
		int pos = output_name.findRev('.');
		if (pos >= 0) {
			int length = output_name.length() - pos - 1;
			return output_name.right(length);
		}
		return output_name;
	};

	int outputChanged() {
		return strcmp((const char *)output_name, (const char *)original_output_name);
	}

	int output_type;
	QString prefix;		// object prefix for the output name;
	QString suffix;		// output name extension: "(final)"|"(intern)"
};

// stack information for equation parser
class QueueInfo
{
public:
	QueueInfo() {
		no = 0;
		value = 0;
		output_id = 0;
		position = 0;
		length = 0;
		type = 0;
	}

	~QueueInfo() { }

	int no;
	QString output_name;
	int value;
	int output_id;
	int type;
	int position;
	int length;
};

// a basic input or output
class Value
{
public:
	Value() {
		internal = false;
		position = 0;
		result = 0;
		undefined_result = 0;
		tristate = false;
	}

	// -----------------------------------
	inline void setValue(int val) {
		result = val;
	}

	inline int getValue() {
		return result;
	}

	// -----------------------------------
	inline void setInternal(bool val) {
		internal = val;
	}

	inline bool isInternal() {
		return internal;
	}

	// -----------------------------------
	inline void setPosition(int newpos) {
		position = newpos;
	}

	inline int getPosition() {
		return position;
	}

	// -----------------------------------
	inline void setUndefined(int _undef) {
		undefined_result = _undef;
	}

	inline int getUndefined() {
		return undefined_result;
	}

	// -----------------------------------
	inline bool isTristate() {
		return tristate;
	}
	 
	inline void setTristate(bool flag) {
		tristate = flag;
	}

protected:
	int undefined_result;
	int result;

private:
	int position;
	bool internal;
	bool tristate;
};

// a basic output with a delayed output value
class DelayedValue : public Value, public Fifo
{
public:
	DelayedValue(int delay) : Value(), Fifo(delay) {
		delayed_result = 0;
	}

	inline void flush(int _value) {
		Fifo::flush(_value);
		result = _value;
		delayed_result = _value;
	}

	// return delayed result
	inline int getResult() {
		return delayed_result;
	}

	// shift result value into fifo, return dropped value
	inline int shift() {
		int ret = put(result);
		delayed_result = get();
		return ret;
	}

	// set new result value
	inline void setValue(int _value) {
		result = _value;
	}

private:
	int delayed_result;
};

#endif

